
/******************************************************************************/
/*  (c)	Copyright Hewlett-Packard Company, 1993,  Fort Collins,	Colorado      */
/*									      */
/*  (C)	COPYRIGHT International	Business Machines Corp.	1993		      */
/*									      */
/*   Copyright (c) 1993	by Sun Microsystems, Inc.			      */
/*									      */
/*			      All Rights Reserved			      */
/*									      */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for any purpose and without fee is hereby granted,	      */
/*  provided that the above copyright notices appear in	all copies and that   */
/*  both the copyright notices and this	permission notice appear in	      */
/*  supporting documentation, and that the names of Hewlett-Packard,	      */
/*  International Business Machines, or	Sun Microsystems not be	used in	      */
/*  advertising	or publicity pertaining	to distribution	of the software	      */
/*  without specific, written prior permission.				      */
/*									      */
/*  HEWLETT-PACKARD MAKES NO WARRANTY OF ANY KIND WITH REGARD TO THIS	      */
/*  SOFTWARE, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF	      */
/*  MERCHANTABILITY AND	FITNESS	FOR A PARTICULAR PURPOSE.  Hewlett-Packard    */
/*  shall not be liable	for errors contained herein or direct, indirect,      */
/*  special, incidental	or consequential damages in connection with the	      */
/*  furnishing,	performance or use of this software.			      */
/*									      */
/*  INTERNATIONAL BUSINESS MACHINES CORPORATION	PROVIDES THIS MATERIAL	      */
/*  "AS	IS" WITHOUT WARRANTY OF	ANY KIND, EITHER EXPRESSED OR IMPLIED,	      */
/*  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY  */
/*  AND	FITNESS	FOR A PARTICULAR PURPOSE.				      */
/*									      */
/*  SUN	MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES WITH REGARD	TO THIS	      */
/*  SOFTWARE, INCLUDING	ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,*/
/*  IN NO EVENT	SHALL SUN MICROSYSTEMS,	INC. BE	LIABLE FOR ANY SPECIAL,	      */
/*  INDIRECT OR	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER	RESULTING     */
/*  FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,      */
/*  NEGLIGENCE OR OTHER	TORTIOUS ACTION, ARISING OUT OF	OR IN CONNECTION      */
/*  WITH THE USE OR PERFORMANCE	OF THIS	SOFTWARE.			      */
/*									      */
/******************************************************************************/

/*
 * This is a SAMPLE program for illustration purposes only.
 * IBM makes no commitment for support of this module.
 */

/*
===============================================================================
	Common Graphics	Environment (CGE) extensions
	PEX 5.1	Simple Programming Example

	Model and scene	generation

===============================================================================
*/
#include <stdio.h>
#include <math.h>

#include <X11/PEX5/PEXlib.h>
#include <X11/PEX5/PEXExtlib.h>

#include "ptorus.h"

/*
*******************************************************************************
    This procedure normalizes a	vector to a magnitude of 1.0.
*******************************************************************************
*/

NormalizeVector( vector	)
    PEXVector *vector;
{
    float r;

    /* Transform a vector to a unit vector. */
    r =	vector->x * vector->x
	+ vector->y * vector->y
	+ vector->z * vector->z;

    if ( r > 0 ) {
      r	= 1.0 /	sqrt( r	);
      vector->x	*= r;
      vector->y	*= r;
      vector->z	*= r;
    }
} /* NormalizeVector */

/*
*******************************************************************************
    This procedure generates the output	commands that constitute the
    model and the closely-associated surface attributes.
*******************************************************************************
*/

void Torus(dpy,	resource_id, req_type)
Display	*dpy;
XID resource_id;
PEXOCRequestType req_type;
{
  /*
   * Torus parameters
   */
    float xaxis				= 10.0,	
    yaxis				= 10.0,	
    zaxis				= 10.0,	
    radius				= 4.0;
    PEXCoord center			= {0,0,0};
    int	depth				= 4; /*	quality	control	*/

    /*
     * Misc variables
     */
    int	i, j;				/* i/phi    j/theta */
    float u, v;
    float phi, theta, step;
    float fphiSize, fthetaSize;
    int	phiSize, thetaSize;
    int	count;
#ifdef TEXTURE
    /* "Custom"	vertex mapping with texture data */
    typedef struct _PEXExtVertexNormal {
      PEXCoord point;
      PEXVector	normal;
      PEXCoord data;
    } PEXExtVertexNormal;
    PEXExtVertexNormal *torus;
    PEXExtArrayOfVertex	quadmesh;
#else
    PEXVertexNormal *torus;
    PEXArrayOfVertex quadmesh;
#endif
    PEXReflectionAttributes ra;
    PEXColor color;
    PEXArrayOfFacetData	facetdata;

    /*
     * Set up a	green object.
     */
    color.rgb.red			= 0.1;
    color.rgb.green			= 0.9;
    color.rgb.blue			= 0.1;
    ra.ambient				= 0.2;
    ra.diffuse				= 1;
    ra.specular				= 1;
    ra.specular_conc			= 15;
    ra.transmission			= 0;
    ra.specular_color.type		= PEXColorTypeRGB;
    ra.specular_color.value.rgb.red	= 0.0402;
    ra.specular_color.value.rgb.green	= 0.4012;
    ra.specular_color.value.rgb.blue	= 0.0402;
    PEXSetSurfaceColor(dpy, resource_id, req_type, PEXColorTypeRGB, &color);
    PEXSetReflectionAttributes(dpy, resource_id, req_type, &ra);
#ifdef TEXTURE
    PEXSetInteriorStyle(dpy, resource_id, req_type, PEXExtInteriorStyleTexture);
#endif

    /*
     * Compute step size and allocate space for	verticies.
     */
    step = (float)(M_PI	/ pow( (double)2, (double)depth	));
    fphiSize = (2 * M_PI / step) + 1.;
    fthetaSize = (2 * M_PI / step) + 1.;
    phiSize = (int)fphiSize;
    thetaSize =	(int)fthetaSize;
#ifdef TEXTURE
    torus = (PEXExtVertexNormal	*) malloc (phiSize * thetaSize * 
					sizeof(PEXExtVertexNormal));
#else
    torus = (PEXVertexNormal *)	malloc (phiSize	* thetaSize * 
					sizeof(PEXVertexNormal));
#endif
    
    /*
     * Sweep from -pi to pi along both u and v to generate verticies.
     */
    count = 0;
    u =	-M_PI -	step;
    for( i = 0;	i < phiSize; i++ ) {
      u	+= step;
      v	= -M_PI	- step;
      for( j = 0; j < thetaSize; j++, count++ )	{
	v +=  step;
	/*
	 * Compute vertex coordinate and normal
	 */
	torus[count].point.x = xaxis * 
	  (radius + cos(v)) * cos(u) + center.x;
	torus[count].point.y = yaxis * 
	  (radius + cos(v)) * sin(u) + center.y;
	torus[count].point.z = zaxis * sin(v) +	center.z;
	torus[count].normal.x =	(1 / xaxis) * cos(v) * cos(u);
	torus[count].normal.y =	(1 / yaxis) * cos(v) * sin(u);
	torus[count].normal.z =	(1 / zaxis) * sin(v);
	NormalizeVector	(&(torus[count].normal));
#ifdef TEXTURE
	torus[count].data.x = (u + M_PI) / (2 *	M_PI);
	torus[count].data.y = (v + M_PI) / (2 *	M_PI);
	torus[count].data.z = 0.0;
#endif
	}
    }


#ifdef TEXTURE
    quadmesh.with_fp_data = (PEXPointer)torus;
    PEXExtQuadrilateralMesh (dpy, 
			     resource_id, 
			     req_type,
			     PEXShapeUnknown, 
			     3,
			     PEXGANone,	
			     PEXGANormal | PEXExtGAData, 
			     PEXColorTypeIndexed,
			     facetdata,	
			     thetaSize,	
			     phiSize, 
			     quadmesh);
#else
    quadmesh.normal = (PEXVertexNormal *)torus;
    PEXQuadrilateralMesh (dpy, 
			  resource_id, 
			  req_type,
			  PEXShapeUnknown, 
			  PEXGANone, 
			  PEXGANormal, 
			  PEXColorTypeIndexed,
			  facetdata, 
			  thetaSize, 
			  phiSize, 
			  quadmesh);
#endif
    free(torus);
}


/*
*******************************************************************************
     This procedure creates a structure	containing the geometric model.
*******************************************************************************
*/

PEXStructure create_model(dpy)
	Display		*dpy;
{
	PEXStructure str_return;
	
	/*
	    Create and load the	structure containing the model.
	*/
	str_return = PEXCreateStructure(dpy);
	Torus (dpy, str_return,	PEXOCStore);

	return str_return;

} /* create_model */


/*
*******************************************************************************
    This procedure creates the light table entries and activation list
    for	the scene.
*******************************************************************************
*/

#define	NUM_LIGHTS 5
PEXLightEntry ltbl[NUM_LIGHTS];
PEXTableIndex lndx[NUM_LIGHTS];

pex_cube_lights(dpy, num_lights, lights, lss)
	Display	*dpy;
	int *num_lights;
	PEXLightEntry **lights;
	PEXTableIndex **lss;
{
	ltbl[0].type = PEXLightAmbient;
	ltbl[0].color.type = PEXColorTypeRGB;
	ltbl[0].color.value.rgb.red   =	0.1;
	ltbl[0].color.value.rgb.green =	0.1;
	ltbl[0].color.value.rgb.blue  =	0.1;
	lndx[0]	= 1;

	ltbl[1].type = PEXLightWCPoint;
	ltbl[1].color.type = PEXColorTypeRGB;
	ltbl[1].color.value.rgb.red   =	0.7653;
	ltbl[1].color.value.rgb.green =	0.763499;
	ltbl[1].color.value.rgb.blue  =	0.7651;
	ltbl[1].point.x	= -19.0;
	ltbl[1].point.y	= 0.0;
	ltbl[1].point.z	= 200.0;
	ltbl[1].attenuation1 = 1;
	ltbl[1].attenuation2 = 0;
	lndx[1]	= 2;

	ltbl[2]	= ltbl[1];
	ltbl[2].point.x	= 17.0;
	ltbl[2].point.y	= 200.0;
	ltbl[2].point.z	= -41.0;
	lndx[2]	= 3;

	ltbl[3]	= ltbl[1];
	ltbl[3].point.x	= 175.0;
	ltbl[3].point.y	= -83.0;
	ltbl[3].point.z	= -51.0;
	lndx[3]	= 4;

	ltbl[4]	= ltbl[1];
	ltbl[4].point.x	= -181.0;
	ltbl[4].point.y	= -83.0;
	ltbl[4].point.z	= 0.0;
	lndx[4]	= 5;

	*num_lights = NUM_LIGHTS;
	*lights	= ltbl;
	*lss = lndx;

} /* pex_cube_lights */


/*
*******************************************************************************
    This procedure creates the viewing and lighting for	the scene.
*******************************************************************************
*/

void create_scene(dpy, camera, num_lights, lights, lss)
	Display	*dpy;
	camera_arg *camera;
	int *num_lights;
	PEXLightEntry **lights;
	PEXTableIndex **lss;
{
	/*
	    Load the camera describing the view.
	*/
	camera->refx = 0.0;
	camera->refy = 0.0;
	camera->refz = 496.0;
	camera->camx = 0.0;
	camera->camy = 0.0;
	camera->camz = 497.0;
	camera->upx = 0.0;
	camera->upy = 1.0;
	camera->upz = 0.0;
	camera->field_of_view =	18.4;
	camera->front =	0.0;
	camera->back = 9999.0;
	camera->perspective = True;

	/*
	    Load the light table entries.
	*/
	pex_cube_lights(dpy, num_lights, lights, lss);

} /* create_scene */


/*
*******************************************************************************
    This procedure sets	up a Pipeline Context with the initial attributes
    required for the scene and model rendering.
*******************************************************************************
*/

void create_initial_state (dpy,	plc, num_lights, lss)
	Display	*dpy;
	PEXPipelineContext plc;
	int num_lights;
	PEXTableIndex *lss;
{
    PEXPCAttributes pc_attrs;
    unsigned long attr_mask[3];

    attr_mask[0] = 0;
    attr_mask[1] = 0;
    attr_mask[2] = 0;

    PEXSetPCAttributeMask(attr_mask, PEXPCCullingMode);
    pc_attrs.culling_mode = PEXBackFaces;

    PEXSetPCAttributeMask(attr_mask, PEXPCDistinguishFlag);
    pc_attrs.distinguish = False;

    PEXSetPCAttributeMask(attr_mask, PEXPCSurfaceEdgeFlag);
    pc_attrs.surface_edges = PEXOff;

    PEXSetPCAttributeMask(attr_mask, PEXPCInteriorStyle);
    pc_attrs.interior_style = PEXInteriorStyleSolid;

    PEXSetPCAttributeMask(attr_mask, PEXPCReflectionModel);
    pc_attrs.reflection_model =	PEXReflectionSpecular;

    PEXSetPCAttributeMask(attr_mask, PEXPCSurfaceInterp);
    pc_attrs.surface_interp = PEXSurfaceInterpColor;

    PEXSetPCAttributeMask(attr_mask, PEXPCViewIndex);
    pc_attrs.view_index	= 1;

    PEXSetPCAttributeMask(attr_mask, PEXPCLightState);
    pc_attrs.light_state.count = num_lights;
    pc_attrs.light_state.indices = lss;

    PEXChangePipelineContext (dpy, plc,	attr_mask, &pc_attrs);

} /* create_initial_state */


/*
*******************************************************************************
    This procedure sets	up the Renderer	and creates the	useful
    associated resources (Lookup Tables	and Pipeline Context).
*******************************************************************************
*/

PEXRenderer create_renderer(dpy, window, rend_attrs)
	Display			*dpy;
	Window			window;
	PEXRendererAttributes	*rend_attrs;
{
    unsigned long itemMask = 0;
    unsigned long attr_mask[3];
    PEXRenderer	renderer;

    itemMask |=	PEXRAPipelineContext;
    attr_mask[0] = 0;
    attr_mask[1] = 0;
    attr_mask[2] = 0;
    rend_attrs->pipeline_context = PEXCreatePipelineContext(dpy, attr_mask,
								NULL);

    itemMask |=	PEXRAViewTable;
    rend_attrs->view_table = PEXCreateLookupTable(dpy, window, PEXLUTView);

    itemMask |=	PEXRALightTable;
    rend_attrs->light_table = PEXCreateLookupTable(dpy,	window,	PEXLUTLight);

    itemMask |=	PEXRAColorApproxTable;
    rend_attrs->color_approx_table = PEXCreateLookupTable(dpy, window, 
							  PEXLUTColorApprox);

    itemMask |=	PEXRABackgroundColor;
    rend_attrs->background_color.type	      =	PEXColorTypeRGB;
    rend_attrs->background_color.value.rgb.red	 = 0.5;
    rend_attrs->background_color.value.rgb.green = 0.5;
    rend_attrs->background_color.value.rgb.blue	 = 0.5;

    itemMask |=	PEXRAHLHSRMode;
    rend_attrs->hlhsr_mode = PEXHLHSRZBuffer;

    itemMask |=	PEXRAClearImage;
    rend_attrs->clear_image = True;

    renderer = PEXCreateRenderer(dpy, window, itemMask,	rend_attrs);
    return renderer;

} /* create_renderer */
